import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import seaborn as sns
from datacyclist.utils import plot_frame



class ActivityStats():
    """
    This class plots all the available stats of an activity when initialized
    The basic plot is a plot of path and the speed (averaged over 10 seconds)
    It also displays a summary of the ride in the form of a table
    
    If power data is available, it will be in a plot (averaged over 20 seconds). Moreover, it will plot
    a pie chart with the time in each power zone, the power curve of this activity and of all time (if available),
    and a histogram with the 25 Watts distribution. At last, the power data will be also in the summary table
    
    If heart rate data is available, it will be in a plot (averaged over 5 seconds). Moreover, it will plot
    a pie chart with the time in each heart rate zone, and the data will be displayed in the summary table.
    
    If cadence data is available, it will be in a plot (averaged over 20 seconds). The data will be also
    displayed in the summary table.
    
    The size of the plot is dynamic, depending on the available data
    
    :param data: pd.Dataframe with columns `power`, `heart_rate`, `cadence`, 
               `speed`, `longitude`, `latitude`, 
                `Power_Training_zone`, `HR_Training_zone`, 
                `distance_covered`.
                
    :param power_curve: PowerCurve object, if provided and power data is present, it will be plotted.
    """
    def __init__(self, data, power_curve=None):
        self.data = data
        self._check_data()
        self.power_curve = power_curve
        self.stat_dict = {'power': 0, 'heart_rate': 0, 'cadence': 0}
        self._update_dict()
        self._make_fig()
        self._plot_path()
        self._plot_speed()
        self._plot_power()
        self._plot_hr()
        self._plot_cadence()
        self._plot_power_curve()
        self._plot_power_dist()
        self._plot_powerzone()
        self._plot_hrzone()
        self._plot_summary()
        plt.show()
        
        
    def _check_data(self):
        diff = {'power', 'heart_rate', 'cadence', 
               'speed', 'longitude', 'latitude', 
                'Power_Training_zone', 'HR_Training_zone', 
                'distance_covered'} - set(self.data.columns)
        if len(list(diff)) > 0:
               raise KeyError(f'Missing columns required: {diff}')
                
        
    def _update_dict(self):
        for stat in ['power', 'heart_rate', 'cadence']:
            if self.data[stat].notna().any():
                self.stat_dict[stat] = 1
        
    def _make_fig(self):
        self.fig = plt.figure(figsize=(15, 10 + 
                                       5 * self.stat_dict['power'] + 
                                       5 * self.stat_dict['heart_rate'] + 
                                       5 * self.stat_dict['cadence'] + 
                                       5 * max(self.stat_dict['power'], self.stat_dict['heart_rate']) ), 
                              facecolor='#292525')
        
        gs = GridSpec(4 + 2*self.stat_dict['power'] + 
                      2*self.stat_dict['heart_rate'] + 
                      2*self.stat_dict['cadence'] + 
                      2*max(self.stat_dict['power'], self.stat_dict['heart_rate']), 4, 
                      figure=self.fig)
        self.path = self.fig.add_subplot(gs[:2, :2])
        self.sum = self.fig.add_subplot(gs[0, 2:])
        self.speed = self.fig.add_subplot(gs[2:4, :])
        if self.stat_dict['power'] == 1:
            self.power_zone = self.fig.add_subplot(gs[1, 2])
            self.power = self.fig.add_subplot(gs[4:6, :])
            if self.power_curve is not None:
                self.power_curve_pl = self.fig.add_subplot(gs[(6+
                                                            2*self.stat_dict['heart_rate']+
                                                            2*self.stat_dict['cadence']):(8
                                                           +2*self.stat_dict['heart_rate']
                                                           +2*self.stat_dict['cadence']), :2])
            self.power_dist = self.fig.add_subplot(gs[(6+
                                                        2*self.stat_dict['heart_rate']+
                                                        2*self.stat_dict['cadence']):(8
                                                       +2*self.stat_dict['heart_rate']
                                                       +2*self.stat_dict['cadence']), 2:])
        if self.stat_dict['heart_rate'] == 1:
            self.hr_zone = self.fig.add_subplot(gs[1, 3])
            self.hr = self.fig.add_subplot(gs[4+2*self.stat_dict['power']:6+2*self.stat_dict['power'], :])
        if self.stat_dict['cadence'] == 1:
            self.cadence = self.fig.add_subplot(gs[(4+
                                                      2*self.stat_dict['power']+
                                                      2*self.stat_dict['heart_rate']):(6+
                                                      2*self.stat_dict['power']+
                                                      2*self.stat_dict['heart_rate']), :])

    def _plot_path(self):
        sns.scatterplot(data=self.data, 
                        x='longitude', y='latitude', c='red', 
                        legend=False,
                        ax=self.path)
        self.path = plot_frame(self.path)
        self.path.set_xlabel('')
        self.path.set_ylabel('')
        self.path.set_xticks([])
        self.path.set_yticks([])
        
        
    def _smooth_series(self, series, kernel_size=10):
#         kernel = np.ones(kernel_size) / kernel_size
#         smooth = np.convolve(series, kernel, mode='same')
        smooth = series.rolling(window=kernel_size, min_periods=kernel_size).mean()
        return smooth
        
    def _plot_speed(self):
        tmp = pd.Series(self._smooth_series(self.data['speed'] *3.6))
        mean_speed = (self.data['speed'] *3.6).mean()
        tmp.plot(ax=self.speed, color='w')
        self.speed.axhline(mean_speed, color='w', linestyle='--', alpha=0.7)
        self.speed = plot_frame(self.speed)
        self.speed.set_xticks([])
        self.speed.set_ylim((0, (self.data['speed'] *3.6).max()+1))
        self.speed.set_title('Speed', fontsize=14, color='w')

        
    def _plot_power(self):
        if self.stat_dict['power'] == 1:
            tmp = pd.Series(self._smooth_series(self.data['power'], kernel_size=20))
            mean_power = self.data['power'].mean()
            tmp.plot(ax=self.power, color='#ECDE15')
            self.power.axhline(mean_power, color='#ECDE15', linestyle='--', alpha=0.7)
            self.power = plot_frame(self.power)
            self.power.set_xticks([])
            self.power.set_ylim((0, (tmp.max()+10)))
            self.power.set_title('Power', fontsize=14, color='w')
            
    def _plot_hr(self):
        if self.stat_dict['heart_rate'] == 1:
            tmp = pd.Series(self._smooth_series(self.data['heart_rate'], kernel_size=5))
            mean_hr = self.data['heart_rate'].mean()
            tmp.plot(ax=self.hr, color='#EC1515')
            self.hr.axhline(mean_hr, color='#EC1515', linestyle='--', alpha=0.7)
            self.hr = plot_frame(self.hr)
            self.hr.set_xticks([])
            self.hr.set_ylim((50, (tmp.max()+10)))
            self.hr.set_title('Heart Rate', fontsize=14, color='w')
            
    def _plot_cadence(self):
        if self.stat_dict['cadence'] == 1:
            tmp = pd.Series(self._smooth_series(self.data['cadence'], kernel_size=20))
            mean_cad = self.data['cadence'].mean()
            tmp.plot(ax=self.cadence, color='#27B012')
            self.cadence.axhline(mean_cad, color='#27B012', linestyle='--', alpha=0.7)
            self.cadence = plot_frame(self.cadence)
            self.cadence.set_xticks([])
            self.cadence.set_ylim((0, (tmp.max()+10)))
            self.cadence.set_title('Cadence', fontsize=14, color='w')
            
    def _plot_power_curve(self):
        if self.stat_dict['power'] == 1 and self.power_curve is not None:
            self.power_curve.calculate_curve(self.data)
            b = self.power_curve.get_activity_curve()
            b['Best Curve'].plot(ax=self.power_curve_pl, label='Best Curve')
            b['Activity Curve'].plot(ax=self.power_curve_pl, label='Activity Curve')
            self.power_curve_pl = plot_frame(self.power_curve_pl)
            self.power_curve_pl.set_xticklabels(['', '1s', '5s', '20s', '1m', '5m', '20m', '2h'])
            self.power_curve_pl.set_title('Power Curve', fontsize=14, color='w')
            self.power_curve_pl.legend(loc='upper right', facecolor='#292525', framealpha=1, labelcolor='w')
            
    def _plot_power_dist(self):
        if self.stat_dict['power'] == 1:
            sns.histplot(data=self.data['power'], binwidth=25, kde=True, ax=self.power_dist)
            self.power_dist = plot_frame(self.power_dist)
            self.power_dist.set_xlabel('')
            self.power_dist.set_ylabel('')
            self.power_dist.set_yticks([])
            self.power_dist.set_title('Power Distribution', fontsize=14, color='w')
            
            
    def _plot_powerzone(self):
        if self.stat_dict['power'] == 1:
            data = self.data['Power_Training_zone'].value_counts().sort_index().values
            wedges, texts = self.power_zone.pie(data, wedgeprops=dict(width=0.5), startangle=-40)
            bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
            kw = dict(arrowprops=dict(arrowstyle="-"),
                      bbox=bbox_props, zorder=0, va="center")
            for i, p in enumerate(wedges):
                ang = (p.theta2 - p.theta1)/2. + p.theta1
                y = np.sin(np.deg2rad(ang))
                x = np.cos(np.deg2rad(ang))
                horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
                connectionstyle = f"angle,angleA=0,angleB={ang}"
                kw["arrowprops"].update({"connectionstyle": connectionstyle})
                self.power_zone.annotate(i+1, xy=(x, y), xytext=(1.3*np.sign(x), 1.2*y),
                            horizontalalignment=horizontalalignment, **kw)
            self.power_zone = plot_frame(self.power_zone)
            self.power_zone.set_title('Power Zones', fontsize=14, color='w')
            
            
    def _plot_hrzone(self):
        if self.stat_dict['heart_rate'] == 1:
            data = self.data['HR_Training_zone'].value_counts().sort_index().values
            wedges, texts = self.hr_zone.pie(data, wedgeprops=dict(width=0.5), startangle=-40)
            bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
            kw = dict(arrowprops=dict(arrowstyle="-"),
                      bbox=bbox_props, zorder=0, va="center")
            for i, p in enumerate(wedges):
                ang = (p.theta2 - p.theta1)/2. + p.theta1
                y = np.sin(np.deg2rad(ang))
                x = np.cos(np.deg2rad(ang))
                horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
                connectionstyle = f"angle,angleA=0,angleB={ang}"
                kw["arrowprops"].update({"connectionstyle": connectionstyle})
                self.hr_zone.annotate(i+1, xy=(x, y), xytext=(1.35*np.sign(x), 1.4*y),
                            horizontalalignment=horizontalalignment, **kw)
            self.hr_zone = plot_frame(self.hr_zone)
            self.hr_zone.set_title('HR Zones', fontsize=14, color='w')
            
            
    def _plot_summary(self):
        data_rows = ['speed'] + self.stat_dict['power'] * ['power'] + self.stat_dict['heart_rate'] * ['heart_rate'] + self.stat_dict['cadence'] * ['cadence']
        rows_names = ['Speed'] + self.stat_dict['power'] * ['Power'] + self.stat_dict['heart_rate'] * ['HR'] + self.stat_dict['cadence'] * ['Cadence'] + ['Distance']
        data = self.data.copy()
        distance = {'mean': '', 'max': round(data.distance_covered.max() / 1000, 2)}
        data['speed'] = data['speed'] * 3.6
        data = round(data[data_rows].agg(['mean', 'max']).T,2 )
        data.loc['Distance'] = distance
        self.sum.table(cellText=data.values, colLabels=['Avg.', 'Max'], loc='center', rowLabels=rows_names, colLoc='center', colWidths=[0.1,0.1])
        self.sum = plot_frame(self.sum)
        self.sum.axis('off')
        self.sum.axis('tight')
        




def plot_frame(ax):
    ax.set_facecolor('#292525')
    ax.spines['bottom'].set_color('w')
    ax.tick_params(axis='x', colors='w')
    ax.xaxis.label.set_color('w')
    ax.spines['left'].set_color('w')
    ax.tick_params(axis='y', colors='w')
    ax.yaxis.label.set_color('w')
    return ax



from datacyclist.powercurve import PowerDict, PowerCurve
from datacyclist.make_dataset import get_dataframes
from datacyclist.fastestsegment import FastestSegment
from datacyclist.plot_activity_summary import ActivityStats
from datacyclist.plot_history import plot_totals, plot_ratios

__all__ = ['PowerDict', 'PowerCurve', 'get_dataframes',
          'FastestSegment',
          'ActivityStats', 'plot_totals', 'plot_ratios']


import datacyclist as dtc
import pandas as pd
import numpy as np
import pytest
from unittest.mock import patch


def create_data():
    df = pd.DataFrame({'activity_no': [1, 1, 1, 1, 1, 1], 
                       'distance': [0,5,5, 5, 5, 1], 
                       'time_elapsed': [0, 1, 2, 3, 4, 5],
                       'time_diff': [1,1,1, 1, 1, 1],
                       'activity_distance': [21]*6, 
                       'distance_covered': [0,5,10,15,20,21], 
                       'year': [1,1,1, 1, 1, 1], 
                       'month': [1,1,1 ,1, 1, 1] })
    return df

df = create_data()


@pytest.mark.parametrize("col", ['activity_no', 'distance', 'time_diff', 'activity_distance', 
                'time_elapsed', 'distance_covered', 'year', 'month'])
def test_missing_columns(col):
    """
    Test if a KeyError is raised if any of the required columns is missing
    """
    data = df.copy()
    del data[col]
    with pytest.raises(KeyError):
        pc = dtc.FastestSegment(data, 1)


def test_toprides_io():
    """
    Test if top_rides is returning a dataframe with the right columns
    """
    fs = dtc.FastestSegment(df, 1)
    top10, monthly = fs.top_rides()
    diff = {'time', 'activity_no', 'year', 'month'} - set(top10.columns)
    assert len(list(diff)) == 0
    diff = {'time', 'activity_no', 'year', 'month'} - set(monthly.columns)
    assert len(list(diff)) == 0
    
    
def test_toprides():
    """
    Test it returns the right values
    """
    tmp = df.copy()
    tmp['activity_no'] = 2
    tmp = pd.concat([df, tmp], ignore_index=True)
    fs = dtc.FastestSegment(tmp, 1)
    top10, monthly = fs.top_rides()
    assert top10['activity_no'].nunique() == 2
    
    
def test_findwindow():
    """
    Test the find_window method, checking the numbers are as expected
    """
    fs = dtc.FastestSegment(df, 20)
    fastest_time, time_start, time_end, activity_no = fs.find_window()
    assert fastest_time == 4
    assert time_start == 1
    assert time_end == 4  
    assert activity_no == 1
    
    
@patch("matplotlib.pyplot.show")
def test_plot_bestrides(_):
    """
    Test if plot_best_rides works
    """
    fs = dtc.FastestSegment(df, 20)
    _, _ = fs.top_rides()
    fs.plot_best_rides()
    
    
@patch("matplotlib.pyplot.show")
def test_plot_bestrides_norides(_):
    """
    Test if plot_best_rides works when best_rides has not been called already
    """
    fs = dtc.FastestSegment(df, 20)
    fs.plot_best_rides()
    



import datacyclist as dtc
import pandas as pd
import numpy as np
import pytest
from unittest.mock import patch


def create_data():
    df = pd.DataFrame({'power': [1,2,4,5]*25, 
                       'speed': [1]*100,
                       'heart_rate': [1]*100,
                       'cadence': [1]*100, 
                       'year': [1]*100,
                       'wk_no': [1]*100, 
                       'month':[1]*100,
                       'distance': [1]*100, 
                       'activity_no': [1]*100})
    return df

df = create_data()


@patch("matplotlib.pyplot.show")
def test_plot_totals(_):
    """
    Test the function works
    """
    dtc.plot_totals(df)

@patch("matplotlib.pyplot.show")
def test_plot_ratios(_):
    """
    Test the function works
    """
    dtc.plot_ratios(df)


import datacyclist as dtc
import pandas as pd
import numpy as np
import pytest
from unittest.mock import patch



def create_data():
    df = pd.DataFrame({'power': [1,2,4,5]*25, 
                       'speed': [1]*100,
                       'heart_rate': [1]*100,
                       'cadence': [1]*100, 
                       'longitude': [1]*100, 
                       'latitude': [1]*100, 
                       'Power_Training_zone': [1]*100,
                       'HR_Training_zone': [1]*100, 
                       'distance_covered': [1]*100})
    return df

df = create_data()

@pytest.mark.parametrize("col", ['power', 'heart_rate', 'cadence', None])
@patch("matplotlib.pyplot.show")
def test_plot_activitystats_nocol_nocurves(_, col):
    """
    Test if the works if non-essential columns are missing
    """
    tmp = df.copy()
    if col is not None:
        tmp[col] = np.nan
    dtc.ActivityStats(tmp, power_curve=None)

    
@pytest.mark.parametrize("col", ['power', 'heart_rate', 'cadence', 
               'speed', 'longitude', 'latitude', 
                'Power_Training_zone', 'HR_Training_zone', 
                'distance_covered'])
def test_missing_columns(col):
    """
    Test if a KeyError is raised if any of the required columns is missing
    """
    data = df.copy()
    del data[col]
    with pytest.raises(KeyError):
        dtc.ActivityStats(data)



import pandas as pd
import numpy as np


class PowerDict(dict):
    def __init__(self):
        self.activity_best = 0
        self.activity_HR = 0
        self.activity_cadence = 0
        self.alltime_best = 0
        self.alltime_HR = 0
        self.alltime_cadence = 0
        
    def __getitem__(self, key):
        return getattr(self, key)
    
    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__dict__[key] = value
        
        
        
class PowerCurve(object):
    """
    This class computes the power curve of an activity and keeps track of the best result for each
    time interval. The method that does the calculation is `calculate_curve` and it is supposed to run over
    the data of 1 activity. It can be run over the data of several activities and to find the best power output
    for a given time interval
    
    It finds the window with the best power output for a given amount of time (from 1 second to 5 hours)
    It also records the corresponding average heart rate and cadence for that time interval (if available)
    It uses the 20 minutes and/or the one hour average power to determine the FTP of the activity and the corresponding
    heart rate (method in development)
    
    :Attributes:
    sec_10: for example, for best results over 10 seconds. This attribute is a dictionary with
    keys `activity best`, `activity_HR`, and `activity_cadence` for the best results of the activity in terms of power 
    and corresponding heart rate and cadence averages (again, over 10 seconds in this example). Other keys are
    `alltime_best` for the best result of all times in terms of 10 seconds power, and `alltime_HR` and `alltime_cadence`
    for their corresponding heart rate and cadence values. These get updated every time the method is rerun over new data.
    
    """
    def __init__(self):
        self.sec_list = [1,2,5,10,20,30,
                         60,120,300,600,1200,
                         3600,7200,18000]
        for sec in self.sec_list:
            setattr(self, f'sec_{sec}', PowerDict())
            
    def _rolling_avgs(self):
        for sec in [sec for sec in self.sec_list if sec <= self.data['time_elapsed'].max()]:
            attr = getattr(self, f'sec_{sec}')
            attr['activity_best'] = 0
            attr['activity_HR'] = 0
            attr['activity_cadence'] = 0
            rolling_avg = self.data.rolling(f'{sec}s', min_periods=sec)['power'].mean()
            rolling_avg_hr = self.data.rolling(f'{sec}s')['heart_rate'].mean()  #todo: not good estimate (maybe with max)
            rolling_avg_cad = self.data.rolling(f'{sec}s')['cadence'].mean()
            attr['activity_best'] = rolling_avg.max()
            max_avg_power_timestamp = rolling_avg.idxmax()
            try:
                attr['activity_HR'] = rolling_avg_hr.loc[max_avg_power_timestamp]
                attr['activity_cadence'] = rolling_avg_cad.loc[max_avg_power_timestamp]
            except KeyError:
                attr['activity_HR'] = 0  # The error is triggered by missing power data or big time gaps
                attr['activity_cadence'] = 0
                #attr['activity_best'] = 0
            if attr['activity_best'] > attr['alltime_best']:
                attr['alltime_best'] = attr['activity_best']
                attr['alltime_HR'] = attr['activity_HR']
                attr['alltime_cadence'] = attr['activity_cadence']
                
    def _check_input(self, data):
        diff = {'timestamp', 'power', 'cadence', 'heart_rate', 'time_elapsed'} - set(data.columns)
        if len(list(diff)) > 0:
            raise KeyError(f'These columns are missing in the data: {list(diff)}')
    
    def calculate_curve(self, activity_data):
        """
        This method finds the best power output over all the time intervals of the curve
        It updates the class attributes.
        """
        self._check_input(activity_data)
        self.data = activity_data.copy().set_index('timestamp')
        self.data = self.data.resample('1S').ffill()
        if self.data['power'].isna().all():
            for sec in self.sec_list:
                attr = getattr(self, f'sec_{sec}')
                attr['activity_best'] = 0
                attr['activity_HR'] = 0
                attr['activity_cadence'] = 0
        else:
            self.data = self.data.fillna(0)
            self._rolling_avgs()            

    def get_activity_curve(self):
        """
        This method returns a dataframe with the values of the power curve, both for the activity and of all time
        """
        activity_curve = []
        best_curve = []
        hr_curve = []
        hr_best = []
        cad_curve = []
        cad_best = []
        for sec in self.sec_list:
            attr = getattr(self, f'sec_{sec}')
            activity_curve.append(attr['activity_best'])
            best_curve.append(attr['alltime_best'])
            hr_curve.append(attr['activity_HR'])
            hr_best.append(attr['alltime_HR'])
            cad_curve.append(attr['activity_cadence'])
            cad_best.append(attr['alltime_cadence'])
        
        res = pd.DataFrame({'Time': self.sec_list, 
                            'Activity Curve': activity_curve, 
                            'Best Curve': best_curve, 
                            'Activity HR': hr_curve, 
                            'HR Best Curve': hr_best, 
                            'Activity Cadence': cad_curve, 
                            'Cadence Best Curve': cad_best})
        return res
    
    
    def get_ftp(self, data):
        """
        This method adds columns to the input data according to the power curve attributes
        It finds the activity FTP and it updates the FTP of the rider if the value improved
        """
        for sec in [1200, 3600]:
            attr = getattr(self, f'sec_{sec}')
            data[f'{int(sec / 60)}min_activity_power'] = np.where(attr['activity_best'] == 0, np.nan, attr['activity_best'])
            data[f'{int(sec / 60)}min_power'] = np.where(attr['alltime_best'] == 0, np.nan, attr['alltime_best'])
            data[f'{int(sec / 60)}min_activity_HR'] = np.where(attr['activity_HR'] == 0, np.nan, attr['activity_HR'])
            data[f'{int(sec / 60)}min_HR'] = np.where(attr['alltime_HR'] == 0, np.nan, attr['alltime_HR'])                      
        
        data['activity_FTP'] = np.where(data['60min_activity_power'].isna(), data['20min_activity_power'] * 0.95,
                                        np.where(data['20min_activity_power'] * 0.95 >= data['60min_activity_power'], 
                                        data['20min_activity_power'] * 0.95, data['60min_activity_power']))
        
        data['FTP'] = np.where(data['60min_power'].isna(), data['20min_power'] * 0.95,
                               np.where(data['20min_power'] * 0.95 >= data['60min_power'], 
                              data['20min_power'] * 0.95, data['60min_power']))
        
        data['FTP_change'] = np.where(data['FTP'] == data['activity_FTP'], 1, 0)
        
        data['Threshold_HR'] = np.where(data['FTP_change'] == 1, 
                                        np.where(data['60min_activity_HR'].isna(), 
                                                 data['20min_activity_HR'] * 0.95,
                                                 np.where(data['20min_activity_power'] * 0.95 >= data['60min_activity_power'], 
                                                          data['20min_activity_HR'] *0.95, data['60min_activity_HR'])),
                                       data['20min_HR']*0.95)
        
        del data['activity_FTP']
        
        data.loc[(data['FTP_change'] == 0) & 
                 ((data['Threshold_HR'] * 0.98 < data['20min_activity_HR'] *0.95) | 
                  (data['Threshold_HR'] * 0.98 < data['60min_activity_HR'])), 'FLAG'] = 1
        data.loc[data['FLAG'] == 1, 'FTP_change'] = 1
        data.loc[data['FLAG'] == 1, 'FTP'] = np.where(data['60min_activity_power'].isna(),
                                                      data['20min_activity_power'] * 0.95,
                                                      np.where(data['20min_activity_power'] * 0.95 >= data['60min_activity_power'], 
                                                               data['20min_activity_power'] * 0.95, 
                                                               data['60min_activity_power']))
#         data.loc[data['FLAG'] == 1, 'Threshold_HR'] = np.where(data['20min_activity_power'] * 0.95 >= data['60min_activity_power'],
#                                                                data['20min_activity_HR'] * 0.95, 
#                                                                data['60min_activity_HR'])
        
        del data['FLAG']
        
        return data



import datacyclist as dtc
import pandas as pd
import numpy as np
import pytest


def create_data():
    df = pd.DataFrame({'timestamp': pd.date_range(start="2023-01-01", periods=11, freq='S'),
                       'time_elapsed': np.arange(0,11),
                       'power': [0,0,1,1,1,1,2,2,2,2,3], 
                       'heart_rate': [0,0,1,1,1,1,2,2,2,2,3], 
                       'cadence': [0,0,1,1,1,1,2,2,2,2,3]})
    return df

df = create_data()

@pytest.mark.parametrize("col", ['timestamp', 'power', 'cadence', 'heart_rate', 'time_elapsed'])
def test_missing_columns(col):
    """
    Test if a KeyError is raised if any of the required columns is missing
    """
    data = df.copy()
    del data[col]
    with pytest.raises(KeyError):
        pc = dtc.PowerCurve()
        pc.calculate_curve(data)


@pytest.mark.parametrize("sec", [1,2,5,10,20,30,60,120,300,600,1200,3600,7200,18000])
def test_curves_attributes(sec):
    """
    Test the attributes are there for each duration and the initialization worked
    """
    pc = dtc.PowerCurve()
    attr = getattr(pc, f'sec_{sec}')
    assert attr['activity_best'] == 0
    assert attr['activity_HR'] == 0
    assert attr['activity_cadence'] == 0
    assert attr['alltime_best'] == 0
    assert attr['alltime_HR'] == 0
    assert attr['alltime_cadence'] == 0
    
    
def test_get_activity_curve():
    """
    Test the method returns a dataframe with the info we need
    """
    pc = dtc.PowerCurve()
    res = pc.get_activity_curve()
    diff = {'Time', 'Activity Curve', 'Best Curve', 'Activity HR', 
            'HR Best Curve', 'Activity Cadence', 'Cadence Best Curve'} - set(res.columns)
    assert len(list(diff)) == 0
    assert res.shape[0] == 14  # the 14 time intervals of the curve


    
def test_curve_calculation():
    """
    Test the results are actually correct
    """
    pc = dtc.PowerCurve()
    pc.calculate_curve(df)
    assert pc.sec_1['activity_best'] == 3
    assert pc.sec_2['activity_best'] == 2.5
    assert pc.sec_5['activity_best'] == 2.2
    assert pc.sec_10['activity_best'] == 1.5
    assert pc.sec_1['activity_HR'] == 3
    assert pc.sec_2['activity_HR'] == 2.5
    assert pc.sec_5['activity_HR'] == 2.2
    assert pc.sec_10['activity_HR'] == 1.5
    assert pc.sec_1['activity_cadence'] == 3
    assert pc.sec_2['activity_cadence'] == 2.5
    assert pc.sec_5['activity_cadence'] == 2.2
    assert pc.sec_10['activity_cadence'] == 1.5
    
    
def test_curve_missing_seconds():
    """
    Test the behavior when a second is missing
    We use forward fill
    """
    data = df.copy()
    data.loc[data.index[3], data.columns] = np.nan
    pc = dtc.PowerCurve()
    pc.calculate_curve(df)
    # Nothing should change
    assert pc.sec_1['activity_best'] == 3
    assert pc.sec_2['activity_best'] == 2.5
    assert pc.sec_5['activity_best'] == 2.2
    assert pc.sec_10['activity_best'] == 1.5
    

def test_ftp():
    df = pd.DataFrame({'timestamp': ['2023-01-01 00:00:00', '2023-01-01 00:20:00', '2023-01-01 01:00:00'],
                   'time_elapsed': [0, 1200, 3600],
                   'power': [200, 200, 300], 
                   'heart_rate': [150, 150, 180], 
                   'cadence': [80, 80, 80]})

    df['timestamp'] = pd.to_datetime(df['timestamp'])
    pc = dtc.PowerCurve()
    pc.calculate_curve(df)
    pc.get_ftp(df)
    
    assert df['FTP_change'].unique()[0] == 1
    assert round(df['FTP'].unique()[0], 0) == 200
    assert round(df['Threshold_HR'].unique()[0], 0) == 150
    
    df['power'] = 300
    pc.calculate_curve(df)
    pc.get_ftp(df)
    assert df['FTP_change'].unique()[0] == 1
    assert round(df['FTP'].unique()[0], 0) == 300
    assert round(df['Threshold_HR'].unique()[0], 0) == 150
    
    df['power'] = 100
    pc.calculate_curve(df)
    pc.get_ftp(df)
    assert df['FTP_change'].unique()[0] == 1
    assert round(df['FTP'].unique()[0], 0) == 100
    assert round(df['Threshold_HR'].unique()[0], 1) == 142.5  #in this case it takes the 20 min power and HR


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from datacyclist.utils import plot_frame



class FastestSegment():
    """
    This class finds the segment in any ride that took the least time. It will find the best time for the segment of
    given lenght in an activity. Thus it won't find multiple segments in the same activity. This is because, for example,
    it is very likely that the activity with the best 30 km segment ever also contains the top 10 of the 30 km segments 
    (a window shifted by just a few seconds is likely to have a very similar time to the one with the best time). 
    This would not be interesting. So an activity can only have 1 segment of 30 km and that will be its best one.
    The method `best_rides` find the top 10 segments of a given distance and the best segment each month.
    The method `plot_best_rides` plots the result of `best_rides`.
    
    :param data: pd.DataFrame, must have the columns `activity_no`, `distance`, `time_diff`, `activity_distance`, 
                `time_elapsed`, `distance_covered`, `year`, `month`.
                
    :param km: length of the segment we want to find the best time of. This is expressed in meters, so calling it
                `km` is an act of pure evil. Note that the smaller is this value, the longer the method will take as
                the candidate segments are much more numerous.
    """
    def __init__(self, data, km):
        self.data = data
        self.km = km
        self._check_data()
        
        
    def _check_data(self):
        diff = {'activity_no', 'distance', 'time_diff', 'activity_distance', 
                'time_elapsed', 'distance_covered', 'year', 'month'} - set(self.data.columns)
        if len(list(diff)) > 0:
            raise KeyError(f'These columns are missing in the data: {list(diff)}')
        
    
    def find_window(self, data=None):
        """
        This method scans each activity to find the best segment of a given distance (the attribute km).
        
        :param data: (optional), a pandas dataframe with the data we want to scan for the best segment.
                            If None, it will use the attribute `data` from the class.
        
        :return fastest_time: the fastest elapsed time for the distance specified by `self.km`
        :return time_start: how many seconds into the ride the segment starts
        :return time_end: how many seconds into the ride the segment ends
        :return activity_no: which activity contains the best segment
        """
        fastest_time = 100000000000000000000
        time_start = 0
        time_end = 0
        activity_no = 0
        
        if data is None:
            data = self.data
        
        for activity in data['activity_no'].unique(): # each activity can only return 1 segment at best
            act_data = data[data['activity_no'] == activity].copy()
            
            new_activity = False # this will be used to break the activity loop if we can't find any other segment
            
            if act_data['distance'].sum() < self.km:
                # only consider activities that are long enough to find a segment of that length
                continue
                
            else:
        
                for i, row in act_data.iterrows():
                    if new_activity: 
                        break
                    start = i
                    time_start = row['time_elapsed']
                    # if the segment is particularly small, this if statement comes to play
                    if self.km <= row['distance'] <= self.km * 1.1:
                        time_diff = row['time_diff']

                        if time_diff < fastest_time:
                            fastest_time = time_diff
                            activity_no = row['activity_no']
                            time_end = time_start + time_diff
                        continue
                    
                    # if the distance remaining in activity is not long enough, stop searching this segment
                    if row['activity_distance'] - row['distance_covered'] < self.km:
                        new_activity = True
                        continue
                    
                    # look for the best time a number of km ahead of the starting point
                    sec_data = act_data[(row['distance_covered'] + self.km <= act_data['distance_covered']) & 
                                        (act_data['distance_covered'] <= row['distance_covered'] + self.km * 1.01)]
                    
                    time_diff = sec_data['time_elapsed'].min() - time_start
                    
                    if time_diff < fastest_time:
                        fastest_time = time_diff
                        activity_no = row['activity_no']
                        time_end = sec_data['time_elapsed'].min()
        
        return fastest_time, time_start, time_end, activity_no
    
    @staticmethod
    def _best_by(x, self):
        best_time, _, _, best = self.find_window(data=x)
        if best_time == 100000000000000000000:
            best_time = np.nan
            best = np.nan
        return best_time, best


    def top_rides(self, n=10):
        """
        This method finds the best rides for the given segment.
        
        :param n: how many top results to return, default=10
        
        :return top10: attribute populated in this method, the top results in the data, from best to worst
        :return by_month: attribute populated in this method, the best result by month, ordered by activity number
        """
        dates = self.data[['activity_no', 'year', 'month']].drop_duplicates()               
        res = self.data.groupby('activity_no').apply(self._best_by, self).reset_index()
        res[['time', 'activity_no']] = pd.DataFrame(res[0].tolist(), index=res.index)
        del res[0]
        res = res.dropna().sort_values(by='time').reset_index(drop=True)
        res['activity_no'] = res['activity_no'].astype(int)
        res['time'] = res['time'].astype(int)
        res = pd.merge(res, dates, on='activity_no', how='left')

        by_month = res.groupby(['year', 'month'], as_index=False)['time'].min()
        by_month = pd.merge(by_month, res, on=['year', 'month', 'time'], how='left')
        by_month = by_month.sort_values(by=['activity_no']).reset_index(drop=True)
        
        self.top10 = res.head(n)
        self.by_month = by_month

        return self.top10, self.by_month


    def plot_best_rides(self):
        """
        Plot the top rides and the top rides by month, side by side
        If the rides have not been computed already, it calls `self.top_rides`
        """
        if hasattr(self, 'top10') and hasattr(self, 'by_month'):
            fig, ax = plt.subplots(1, 2, figsize=(15, 6), facecolor='#292525')
            fig.suptitle(f'Best rides of {int(self.km/1000)} Km', fontsize=18, color='w')
            (self.top10.sort_values(by='activity_no').set_index(['year', 'month'])['time'] / 3600).plot(ax=ax[0], color='#46a832')
            (self.by_month.set_index(['year', 'month'])['time'] / 3600).plot(ax=ax[1], color='#a86b32')
            ax[0] = plot_frame(ax[0])
            ax[1] = plot_frame(ax[1])
            ax[0].set_xlabel('')
            ax[1].set_xlabel('')
            ax[0].set_title('Top 10 rides', color='w', fontsize=14)
            ax[1].set_title('Best ride by month', color='w', fontsize=14)
            plt.show()
        else:
            _, _ = self.top_rides()
            self.plot_best_rides()


#! /usr/bin/env python

DESCRIPTION = "datacyclist, a package to analyze cycling data."
with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()

DISTNAME = 'datacyclist'
MAINTAINER = 'Luca Basanisi'
MAINTAINER_EMAIL = 'luca.basanisi@gmail.com'
URL = 'https://pypi.org/project/datacyclist/'
LICENSE = 'Apache 2.0'
DOWNLOAD_URL = 'https://github.com/lucabasa/datacyclist'
PROJECT_URLS = {
    #'Bug Tracker': ,
    #'Documentation': 'TBD',
    'Source Code': 'https://github.com/lucabasa/datacyclist'
}
VERSION = '0.2.0'
PYTHON_REQUIRES = ">=3.7"

INSTALL_REQUIRES = [
    'matplotlib>=3.5.3',
    'pandas>=1.3.5',
    'scikit-learn>=1.0.2',
    'seaborn>=0.12.0',
    'fitdecode==0.10.0'
]


PACKAGES = [
    'datacyclist',
]

CLASSIFIERS = [
    'Development Status :: 4 - Beta', 
    'Programming Language :: Python', 
    'Programming Language :: Python :: 3.10', 
    'Intended Audience :: Science/Research',
    'Intended Audience :: Developers',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Operating System :: MacOS',
]

TEST_REQUIRE = [
        "pytest==7.1.3", 
        "coverage==6.4.4", 
        'xgboost==1.6.2',
        'lightgbm==3.3.2'
]



if __name__ == "__main__":

    from setuptools import setup

    import sys
    if sys.version_info[:2] < (3, 7):
        raise RuntimeError("datacyclist requires python >= 3.7.")

    setup(
        name=DISTNAME,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        license=LICENSE,
        url=URL,
        test_require=TEST_REQUIRE,
        version=VERSION,
        download_url=DOWNLOAD_URL,
        project_urls=PROJECT_URLS,
        python_requires=PYTHON_REQUIRES,
        install_requires=INSTALL_REQUIRES,
        packages=PACKAGES,
        classifiers=CLASSIFIERS
    )



import fitdecode

import pandas as pd
from datetime import datetime, timedelta
from typing import Dict, Union, Optional,Tuple

#thanks to this script https://github.com/bunburya/fitness_tracker_data_parsing/blob/main/parse_fit.py

POINTS_COLUMN_NAMES = ['latitude', 'longitude', 'lap', 'altitude', 'timestamp', 'heart_rate', 
                       'cadence', 'speed', 'power','temperature', 'left_power_phase', 'left_power_phase_peak']

# The names of the columns we will use in our laps DataFrame. 
LAPS_COLUMN_NAMES = ['lap', 'start_time', 'total_distance', 'total_elapsed_time', 
                     'total_strokes', 'total_work', 'total_calories', 'time_standing',
                     'avg_speed', 'max_speed',
                     'avg_power', 'max_power',
                     'total_ascent', 'total_descent',
                     'avg_cadence', 'max_cadence',
                     'avg_temperature', 'max_temperature',
                     'normalized_power', 
                     'max_heart_rate', 'avg_heart_rate', 
                     'avg_left_power_phase', 'avg_left_power_phase_peak']


def get_fit_lap_data(frame: fitdecode.records.FitDataMessage) -> Dict[str, Union[float, datetime, timedelta, int]]:
    """Extract some data from a FIT frame representing a lap and return
    it as a dict.
    """
    
    data: Dict[str, Union[float, datetime, timedelta, int]] = {}
    
    for field in LAPS_COLUMN_NAMES[1:]:  # Exclude 'number' (lap number) because we don't get that
                                        # from the data but rather count it ourselves
        if frame.has_field(field):
            data[field] = frame.get_value(field)
    
    return data


def get_fit_point_data(frame: fitdecode.records.FitDataMessage) -> Optional[Dict[str, Union[float, int, str, datetime]]]:
    """Extract some data from an FIT frame representing a track point
    and return it as a dict.
    """
    
    data: Dict[str, Union[float, int, str, datetime]] = {}
    
    if not (frame.has_field('position_lat') and frame.has_field('position_long')):
        # Frame does not have any latitude or longitude data. We will ignore these frames in order to keep things
        # simple, as we did when parsing the TCX file.
        return None
    else:
        data['latitude'] = frame.get_value('position_lat') / ((2**32) / 360)
        data['longitude'] = frame.get_value('position_long') / ((2**32) / 360)
    
    for field in POINTS_COLUMN_NAMES[3:]:
        if frame.has_field(field):
            data[field] = frame.get_value(field)
    
    return data


def crop_activity(data, minutes=10):
    activity_start = data['timestamp'].min()
    activity_end = data['timestamp'].max()
    
    cropped = data.loc[(data['timestamp'] > activity_start + pd.Timedelta(minutes=minutes)) & 
                       (data['timestamp'] < activity_end - pd.Timedelta(minutes=minutes)), ]
    
    return cropped.reset_index(drop=True)
    

def get_dataframes(fname: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Takes the path to a FIT file (as a string) and returns two Pandas
    DataFrames: one containing data about the laps, and one containing
    data about the individual points.
    """

    points_data = []
    laps_data = []
    lap_no = 1
    with fitdecode.FitReader(fname) as fit_file:
        for frame in fit_file:
            if isinstance(frame, fitdecode.records.FitDataMessage):
                if frame.name == 'record':
                    single_point_data = get_fit_point_data(frame)
                    if single_point_data is not None:
                        single_point_data['lap'] = lap_no
                        points_data.append(single_point_data)
                elif frame.name == 'lap':
                    single_lap_data = get_fit_lap_data(frame)
                    single_lap_data['lap'] = lap_no
                    laps_data.append(single_lap_data)
                    lap_no += 1
    
    # Create DataFrames from the data we have collected. If any information is missing from a particular lap or track
    # point, it will show up as a null value or "NaN" in the DataFrame.
    
    laps_df = pd.DataFrame(laps_data, columns=LAPS_COLUMN_NAMES)
    points_df = pd.DataFrame(points_data, columns=POINTS_COLUMN_NAMES)
    
    # more privacy
    points_df = crop_activity(points_df)
    
    return laps_df, points_df






import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from datacyclist.utils import plot_frame


def plot_totals(data):
    """
    This function plots the total distance, average power, speed, and cadence per week.
    
    :param data: pd.DataFrame with distance, power, speed, and cadence data over time
    """
    fig, ax = plt.subplots(4, 1, figsize=(15, 20), facecolor='#292525')
    fig.subplots_adjust(top=0.95)
    fig.suptitle(f' Analysis', fontsize=18, color='w')
    
    (data.groupby(['year', 'wk_no'])['distance'].sum() / 1000).plot(ax=ax[0], kind='bar')
    data.groupby(['year', 'wk_no'])['power'].mean().plot(kind='bar', ax=ax[1], color='#ECDE15')
    (data.groupby(['year', 'wk_no'])['speed'].mean()*3.6).plot(kind='bar', ax=ax[2])
    data.groupby(['year', 'wk_no'])['cadence'].mean().plot(kind='bar', ax=ax[3], color='#27B012')
        
    ax[0].set_title('Distance', fontsize=14, color='w')
    ax[1].set_title('Power', fontsize=14, color='w')
    ax[2].set_title('Speed', fontsize=14, color='w')
    ax[3].set_title('Cadence', fontsize=14, color='w')
    
    for axes in ax:
        axes = plot_frame(axes)
        axes.set_xlabel('')
        axes.set_xticks([])
        
    plt.show()

    
def plot_ratios(data):
    """
    This function plots a series of scatter plots of mean power, cadence, speed and heart rate per activity
    For example, mean speed vs mean power
    
    :param data: pd.DataFrame with data of power, cadence, speed, and heart rate per activity
    """
    fig, ax = plt.subplots(3, 2, figsize=(15, 15), facecolor='#292525')
    fig.subplots_adjust(top=0.95)
    fig.suptitle(f'Mean Ratios by activity', fontsize=18, color='w')
    
    df = data.groupby('activity_no', as_index=False)[['power', 'cadence', 'speed', 'heart_rate']].mean()
    df['speed'] = df['speed'] * 3.6
    dates = data[['activity_no', 'year', 'month']].drop_duplicates()
    df = pd.merge(df, dates, on='activity_no', how='left')
        
    sns.scatterplot(df, x='power', y='speed', ax=ax[0][0])
    sns.scatterplot(df, x='power', y='cadence', ax=ax[0][1])
    sns.scatterplot(df, x='power', y='heart_rate', ax=ax[1][0])
    sns.scatterplot(df, x='cadence', y='heart_rate', ax=ax[1][1])
    sns.scatterplot(df, x='cadence', y='speed', ax=ax[2][0])
    sns.scatterplot(df, x='heart_rate', y='speed', ax=ax[2][1])

    ax[0][0] = plot_frame(ax[0][0])
    ax[0][1] = plot_frame(ax[0][1])
    ax[1][0] = plot_frame(ax[1][0])
    ax[1][1] = plot_frame(ax[1][1])
    ax[2][0] = plot_frame(ax[2][0])
    ax[2][1] = plot_frame(ax[2][1])
        
    plt.show()
